#include <algorithm>    // for std::find

#include "Node.h"
#include "Log.h"		// for LOG_DEBUG(...)
#include "NodeView.h"
#include "Wire.h"
#include "WireView.h"
#include "History.h"
#include "DataAccess.h"
#include "ComputeBase.h"
#include "NodeTraversal.h"

using namespace Nodable;

void Node::Disconnect(Wire* _wire)
{
	_wire->getTarget()->setInputMember(nullptr);

	auto targetNode = _wire->getTarget()->getOwner()->as<Node>();	
	auto sourceNode = _wire->getSource()->getOwner()->as<Node>();

	targetNode->removeWire(_wire);
	sourceNode->removeWire(_wire);

	NodeTraversal::SetDirty(targetNode);

    delete _wire;

	return;
}

Wire* Node::Connect( Member* _from, Member* _to)
{	
	Cmd_ConnectWire command(_from, _to);
	command.execute();
	return command.getWire();
}

Node::Node(std::string _label):

	parentContainer(nullptr),
	label(_label),
	dirty(true)
{

}

Node::~Node()
{
    // Disconnect and clear wires
    std::for_each(wires.crbegin(), wires.crend(), [](auto item) {
       Node::Disconnect(item);
    });

	// Delete all components
	for(auto pair : components)
	{
		delete pair.second;
	}

}

bool Node::isDirty()const
{
	return dirty;
}

void Node::setDirty(bool _value)
{
	dirty = _value;
}

Container* Node::getParentContainer()const
{
	return this->parentContainer;
}

void Node::setParentContainer(Container* _container)
{
	this->parentContainer = _container;
}

void Node::setLabel(const char* _label)
{
	this->label = _label;
}

void Node::setLabel(std::string _label)
{
	this->label = _label;
}

const char* Node::getLabel()const
{
	return this->label.c_str();
}

void Nodable::Node::addWire(Wire* _wire)
{
	wires.push_back(_wire);
}

void Nodable::Node::removeWire(Wire* _wire)
{
	auto found = std::find(wires.begin(), wires.end(), _wire);
	if(found != wires.end())
		wires.erase(found);
}

std::vector<Wire*>& Node::getWires()
{
	return wires;
}

int Node::getInputWireCount()const
{
	int count = 0;
	for(auto w : wires)
	{
		if ( w->getTarget()->getOwner() == this)
			count++;
	}
	return count;
}

int Node::getOutputWireCount()const
{
	int count = 0;
	for(auto w : wires)
	{
		if ( w->getSource()->getOwner() == this)
			count++;
	}
	return count;
}

UpdateResult Node::update()
{
    // TODO: take in account the result of component's update()

	if(hasComponent<ComputeBase>())
    {
        getComponent<ComputeBase>()->update();
    }

	if(hasComponent<DataAccess>())
    {
        getComponent<DataAccess>()->update();
    }

	return UpdateResult::SuccessWithChanges;
}

void Node::onMemberValueChanged(const char* _name)
{	
	updateLabel();
	NodeTraversal::SetDirty(this);
}